/***************************************************************************
 * Sonic Retro Image Recovery Program v0.4                                 *
 *                                                                         *
 * Copyright (c) 2009 by David Korth.                                      *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.           *
 ***************************************************************************/

/**
 * Changelog:
 * - 0.4: Use lstat() to skip symlinks; support Unicode build on Win32.
 * - 0.3: Close directories after using them.
 * - 0.2: Don't error out if a directory can't be opened.
 * - 0.1: Initial public release.
 */

#ifdef _WIN32
#define DIRSEP '\\'
#else
#define DIRSEP '/'
#endif

#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

// tchar.h is required for Unicode on Win32.
#include "rf_tchar.h"

// MinGW doesn't support S_ISLNK.
#ifndef S_ISLNK
#define S_ISLNK(x) 0
#endif

#define BUFFER_SIZE 1048576

#include <string>
#include <list>
using std::string;
using std::list;

#include <mhash.h>

// TR1 unordered map.
#include <tr1/unordered_map>
#include <utility>
using std::tr1::unordered_map;

// SHA1 to Filename map.
typedef unordered_map<string, string> map_SHA1_t;
map_SHA1_t map_SHA1;
typedef std::pair<string, string> pair_SHA1;

// MATCH.log
FILE *f_match = NULL;

// Number of files found.
int num_files = 0;

string get_sha1(const TCHAR* filename)
{
	// Get the SHA1 of the specified file.
	FILE *f = _tfopen(filename, TEXT("rb"));
	if (!f)
	{
		_ftprintf(stderr, TEXT("Error opening file '%s'.\n"), filename);
		printf("Error: %d: %s\n", errno, strerror(errno));
		return "";
	}
	
	MHASH td;
	unsigned char buffer[BUFFER_SIZE];
	unsigned char *hash;
	int i;
	
	td = mhash_init(MHASH_SHA1);
	while (!feof(f))
	{
		i = fread(buffer, 1, sizeof(buffer), f);
		if (i == 0)
			break;
		mhash(td, &buffer, i);
	}
	fclose(f);
	hash = (unsigned char*)mhash_end(td);
	
	// Convert the SHA1 to hexadecimal ASCII.
	char digest_string[64];
	char *digest_strptr = &digest_string[0];
	int digest_strlen = mhash_get_block_size(MHASH_SHA1);
	for (i = 0; i < digest_strlen; i++)
	{
		*digest_strptr = (hash[i] >> 4) | 0x30;
		if (*digest_strptr >= 0x3A)
			*digest_strptr += 0x27;
		digest_strptr++;
		
		*digest_strptr = (hash[i] & 0x0F) | 0x30;
		if (*digest_strptr >= 0x3A)
			*digest_strptr += 0x27;
		digest_strptr++;
	}
	*digest_strptr = 0x00;
	
	return string(digest_string);
}

void search_directory(const TCHAR* dir_name)
{
	_TDIR *dir = _topendir(dir_name);
	if (!dir)
	{
		_ftprintf(stderr, TEXT("Error opening directory '%s'.\n"), dir_name);
		return;
	}
	_ftprintf(stderr, TEXT("Scanning: %s\n"), dir_name);
	
	struct _tdirent *d_entry;
#ifdef WIN32
	struct _stat d_stat;
#else
	struct stat d_stat;
#endif
	
	// Check subdirectories later.
	list<CPPstring> subdirs;
	
	map_SHA1_t::iterator sha1_iter;
	
	// Check every file.
	while ((d_entry = _treaddir(dir)))
	{
		TCHAR full_path[1024];
		_sntprintf(full_path, sizeof(full_path)/sizeof(TCHAR),
			   TEXT("%s%c%s"), dir_name, DIRSEP, d_entry->d_name);
#ifdef _WIN32
		_tstat(full_path, &d_stat);
#else
		lstat(full_path, &d_stat);
#endif
		
		if (S_ISLNK(d_stat.st_mode))
		{
			// Skip symbolic links.
			continue;
		}
		else if (S_ISDIR(d_stat.st_mode))
		{
			// Directory.
			if (_tcscmp(d_entry->d_name, TEXT(".")) != 0 &&
			    _tcscmp(d_entry->d_name, TEXT("..")) != 0)
			{
				// Not "." or "..". Scan it later.
				subdirs.push_back(CPPstring(full_path));
			}
		}
		else if (S_ISREG(d_stat.st_mode))
		{
			// Regular file. Get its SHA1.
			string file_SHA1 = get_sha1(full_path);
			
			// Check if it matches anything.
			sha1_iter = map_SHA1.find(file_SHA1);
			if (sha1_iter != map_SHA1.end())
			{
				// Found a match!
				_tprintf(TEXT("MATCH: %s\n"), full_path);
				_ftprintf(f_match, TEXT("%s\n"), full_path);
				fflush(f_match);
				
				// Increment the file counter.
				num_files++;
			}
		}
	}
	
	// Close the directory.
	_tclosedir(dir);
	
	// Check subdirectories.
	for (list<CPPstring>::iterator iter = subdirs.begin();
	     iter != subdirs.end(); iter++)
	{
		search_directory((*iter).c_str());
	}
}

int load_data_file(void)
{
	// Load "retrofix.dat".
	FILE *fixdat = _tfopen(TEXT("retrofix.dat"), TEXT("r"));
	if (!fixdat)
	{
		_ftprintf(stderr, TEXT("Error: Could not open retrofix.dat.\n"));
		return -1;
	}
	
	// Read all the lines.
	// NOTE: retrofix.dat is in ANSI, not Unicode.
	char cur_line[1024];
	string sha1, filename;
	while (!feof(fixdat))
	{
		fgets(cur_line, sizeof(cur_line), fixdat);
		
		// NULL out any CRs or LFs.
		for (unsigned int i = 0; i < sizeof(cur_line); i++)
		{
			if (!cur_line[i] || cur_line[i] == '\n' || cur_line[i] == '\r')
			{
				cur_line[i] = 0;
				break;
			}
		}
		
		// If the first character is '#' or NULL, ignore the line.
		if (!cur_line[0] || cur_line[0] == '#')
			continue;
		
		// If the line is less than 43 characters (40 SHA1 + filename), ignore it.
		if (strlen(cur_line) < 43)
			continue;
		
		// Separate the line.
		cur_line[40] = 0x00;
		sha1 = string(cur_line);
		filename = string(&cur_line[42]);
		
		// Add the file to the hash map.
		map_SHA1.insert(pair_SHA1(sha1, filename));
	}
	fclose(fixdat);
	
	return 0;
}

void sighandler(int signum)
{
	_ftprintf(stderr, TEXT("Signal %d received. Shutting down.\n"), signum);
	if (f_match)
		fclose(f_match);
	exit(signum);
}

int main(int argc, char** argv)
{
	static const TCHAR *retrofix_intro = TEXT(
		"Sonic Retro Image Recovery Program v0.4\n"
		"Copyright 2009 by David Korth.\n"
		"This program is licensed under the GNU General Public License v2.\n"
		"See http://www.gnu.org/licenses/gpl-2.0.html for more information.\n\n"
		);
	_ftprintf(stderr, TEXT("%s"), retrofix_intro);
	
	// Set up the signal handler.
	signal(SIGINT, sighandler);
	signal(SIGFPE, sighandler);
	signal(SIGABRT, sighandler);
	signal(SIGTERM, sighandler);
	signal(SIGSEGV, sighandler);
	signal(SIGILL, sighandler);
#ifdef SIGQUIT
	signal(SIGQUIT, sighandler);
#endif
#ifdef SIGHUP
	signal(SIGHUP, SIG_IGN);
#endif
	
	// Load the data file.
	int rval = load_data_file();
	if (rval)
		return rval;
	
	// Open the match log file.
	f_match = _tfopen(TEXT("MATCH.log"), TEXT("w"));
	if (!f_match)
	{
		_ftprintf(stderr, TEXT("Error: Could not open MATCH.log for writing.\n"));
		return -1;
	}
	
#if defined(_WIN32) && (defined(UNICODE) || defined(_UNICODE))
	// Win32 Unicode. Print the Unicode byte-order mark.
	_fputtc(0xFEFF, f_match);
#endif

	_ftprintf(f_match, TEXT("%s"), retrofix_intro);
	_ftprintf(stderr, TEXT("Matching files will be written to MATCH.log.\n"));
	
	_ftprintf(stderr, TEXT("Recursively scanning the current directory for matches...\n\n"));
	_ftprintf(f_match, TEXT("List of files matching the Sonic Retro wiki database:\n\n"));
	search_directory(TEXT("."));
	
	_ftprintf(stderr, TEXT("\nSearch completed. %d matching file%s found.\n"),
			num_files, (num_files == 1 ? TEXT("") : TEXT("s")));
	_ftprintf(f_match, TEXT("\nSearch completed. %d matching file%s found.\n"),
			num_files, (num_files == 1 ? TEXT("") : TEXT("s")));
	
	fclose(f_match);
	return 0;
}
